/*
    Copyright (C) 2000-2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: region.h,v 1.31 2003/10/29 04:27:23 pbd Exp $
*/

#ifndef __ardour_region_h__
#define __ardour_region_h__

#include <pbd/undo.h>

#include <ardour/ardour.h>
#include <ardour/logcurve.h>

class XMLNode;

namespace ARDOUR {

class Playlist;

struct RegionState {

    RegionState() {}
    
    jack_nframes_t      _start;
    jack_nframes_t      _length;
    jack_nframes_t      _position;
    unsigned short 	_flags;
    jack_nframes_t      _sync_offset;
    layer_t        	_layer;
    string              _name;        
    mutable bool        _first_edit;
};

class Region : public Stateful, public SigC::Object 
{
  public:
	enum Flag {
		Muted = 0x1,
		Opaque = 0x2,
		EnvelopeActive = 0x4,
		DefaultFadeIn = 0x8,
		DefaultFadeOut = 0x10,
		Locked = 0x20,
		Automatic = 0x40,
		WholeFile = 0x80,
		FadeIn = 0x100,
		FadeOut = 0x200,
		Copied = 0x400,
		Import = 0x800,
		External = 0x1000
	};
	
	static const Flag DefaultFlags = Flag (Opaque|DefaultFadeIn|DefaultFadeOut|FadeIn|FadeOut);
	
        // don't extend this without also working on FreezeIndex below
	
	enum FreezeFlag {
		BoundsNotification = 0x1,
		FadeNotification = 0x2,
		SyncNotification = 0x8,
		EnvelopeNotification = 0x10,
		LayerNotification = 0x20
	};

	Region (jack_nframes_t start, jack_nframes_t length, 
		const string& name, layer_t = 0, Flag flags = DefaultFlags);
	Region (const Region&, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t = 0, Flag flags = DefaultFlags);
	Region (const Region&, const string& name);
	Region (const Region&);
	Region (const XMLNode&);
	~Region();

	ARDOUR::id_t id() const { return _id; }

	/* Note: changing the name of a Region does not constitute an edit */

	string name() const { return _name; }
	void set_name (string str);

	jack_nframes_t position () const { return _position; }
	jack_nframes_t start () const { return _start; }
	jack_nframes_t length() const { return _length; }
	jack_nframes_t sync_offset() const { return _sync_offset; }
	layer_t layer () const { return _layer; }
	
	/* first_frame() is an alias; last_frame() just hides some math */

	jack_nframes_t first_frame() const { return _position; }
	jack_nframes_t last_frame() const { return _position + _length - 1; }

	bool muted() const { return _flags & Muted; }
	bool opaque () const { return _flags & Opaque; }
	bool envelope_active () const { return _flags & EnvelopeActive; }
	bool locked() const { return _flags & Locked; }
	bool automatic() const { return  _flags & Automatic; }
	bool whole_file() const { return _flags & WholeFile ; }

	void freeze (FreezeFlag);
	void thaw (FreezeFlag, const string& why, bool with_sig, void *src);

	bool covers (jack_nframes_t frame) const {
		return _position <= frame && frame < _position + _length;
	}

	enum OverlapType {
		OverlapNone,
		OverlapInternal,
		OverlapStart,
		OverlapEnd,
		OverlapExternal,
	};

	OverlapType coverage (jack_nframes_t start, jack_nframes_t end) {
		jack_nframes_t rend = _position + _length - 1;

		if ((start >= _position) && (end <= rend)) {
			return OverlapInternal;
		}
		if ((end >= _position) && (end <= rend)) {
			return OverlapStart;
		}
		if ((start >= _position) && (start <= rend)) {
			return OverlapEnd;
		}
		if ((_position >= start) && (_position <= end) && (rend <= end)) {
			return OverlapExternal;
		}
		return OverlapNone;
	}

	virtual bool is_crossfade() const { return false; }

	virtual jack_nframes_t read_at (Sample *buf, Sample *mixdown_buffer, 
					float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n=0) const = 0;

	SigC::Signal2<void, BoundsChangeInfo,void*> BoundsChanged;
	SigC::Signal0<void> NameChanged;
	SigC::Signal0<void> MuteChanged;
	SigC::Signal0<void> OpacityChanged;
	SigC::Signal0<void> SyncOffsetChanged;
	SigC::Signal0<void> LockChanged;
	SigC::Signal0<void> LayerChanged;
	SigC::Signal1<void,void*> SelectedChanged;

	XMLNode& get_state ();
	int      set_state (const XMLNode&);
	XMLNode& get_short_state (); /* used only by Session */

	/* EDITING OPERATIONS */

	void set_length (jack_nframes_t, void *src);
	void set_start (jack_nframes_t, void *src);
	void set_position (jack_nframes_t, void *src);
	void set_position_on_top (jack_nframes_t, void *src);

	void nudge_position (long, void *src);

	void trim_start (jack_nframes_t new_position, void *src);
	void trim_front (jack_nframes_t new_position, void *src);
	void trim_end (jack_nframes_t new_position, void *src);
	void trim_to (jack_nframes_t position, jack_nframes_t length, void *src);
	
	void set_layer (layer_t l); /* ONLY Playlist can call this */
	void raise ();
	void lower ();
	void raise_to_top ();
	void lower_to_bottom ();

	void set_sync_offset (jack_nframes_t n);
	void set_muted (bool yn);
	void set_opaque (bool yn);
	void set_envelope_active (bool yn);
	void set_locked (bool yn);

	virtual unsigned long read_data_count() const { return _read_data_count; }
	virtual void set_state (state_id_t) = 0;

	ARDOUR::Playlist* playlist() const { return _playlist; }

	UndoAction get_memento() const;

  private:
	friend class Playlist;
	virtual void save_state (const string &) = 0;
	
  protected:

	enum FreezeIndex {
		BoundsNotificationIndex,
		FadeNotificationIndex,
		SyncNotificationIndex,
		EnvelopeNotificationIndex,
		LayerNotificationIndex,
		MaxFrozen // must be last
	};

	jack_nframes_t        _start;
	jack_nframes_t        _length;
	jack_nframes_t        _position;
	unsigned short 	      _flags;
	jack_nframes_t        _sync_offset;
	layer_t        	      _layer;
	string                _name;        
	mutable bool          _first_edit;
	unsigned char         _frozen[MaxFrozen];
	PBD::Lock              lock;
	ARDOUR::id_t          _id;
	state_id_t            _current_state_id;
	ARDOUR::Playlist*     _playlist;
	mutable unsigned long _read_data_count; // modified in read()

	bool hold_bounds_notification() const { return _frozen[BoundsNotificationIndex]; }
	bool hold_sync_notification() const { return _frozen[SyncNotificationIndex]; }
	bool hold_layer_notification() const { return _frozen[LayerNotificationIndex]; }
	
	virtual void store_state (RegionState&) const;

	/* this is not virtual because of the complexity of the logic. 
	   instead it calls check_restore_flags() midway through to 
	   give derived classes a chance to do their thing.
	*/

	void restore_state (RegionState&);

	XMLNode& _get_state (bool is_short);
	
	void trim_to_internal (jack_nframes_t position, jack_nframes_t length, void *src);
	void set_playlist (ARDOUR::Playlist*);

	bool copied() const { return _flags & Copied; }
	void maybe_uncopy ();
	void rename_after_first_edit ();
	
	virtual bool verify_start (jack_nframes_t) = 0;
	virtual bool verify_start_and_length (jack_nframes_t, jack_nframes_t) = 0;
	virtual bool verify_start_mutable (jack_nframes_t&_start) = 0;
	virtual bool verify_length (jack_nframes_t) = 0;
	virtual void recompute_at_start () = 0;
	virtual void recompute_at_end () = 0;
	virtual void check_restore_flags (RegionState& state, FreezeFlag& flags) = 0;
	virtual void xtra_freeze (FreezeFlag) = 0;
	virtual void xtra_thaw (FreezeFlag, FreezeFlag&) = 0;
	virtual void thaw_signals (FreezeFlag) = 0;
};

} /* namespace ARDOUR */

#endif /* __ardour_region_h__ */
